/*+*****************************************************************************
*                                                                              *
* File: auth.c                                                                 *
*                                                                              *
* Description: X authorization handling                                        *
*                                                                              *
**-****************************************************************************/

#ifndef __lint
char auth_vers[] = "$Id: auth.c,v 1.11 2008/07/03 12:25:17 cons Exp $";
#endif /* __lint */

/******* Include Files ********************************************************/

#include "auth.h"
#include "util.h"
#include "io.h"
#include "host.h"
#include "mc.h"
#include "xmd.h"

#include <netinet/in.h>
#include <X11/X.h>
#include <X11/Xauth.h>

/******* Constants ************************************************************/

/******* Macros ***************************************************************/

/******* External Stuff *******************************************************/

/******* Local Stuff **********************************************************/

static Xauth *auth_data = NULL;	/* static copy of the authorization data */

/*
 * try to find the authorization data for the given family and display
 */
int auth_try(int family, HOST *shost, int idisplay)
{
  int attempt, saddrlen, dpynumlen;
  char dpynumbuf[8], *saddr, *fname;

  /* do nothing if we already have authorization data */
  if (auth_data) return(-1);

  /* check the family */
  fname = NULL;
  switch (family) {
  case FamilyLocal:     fname = "Local";     break;
  case FamilyLocalHost: fname = "LocalHost"; break;
  case FamilyInternet:  fname = "Internet";  break;
  default: die("unexpected authorization family: %d", family);
  }

  /* setup */
#ifdef DEBUG
  if (DEBUGGING(DBG_MISC))
    dprintf("auth", "looking for %s authorization for %s:%d",
	    fname, shost->fq_name, idisplay);
#endif
  sprintf(dpynumbuf, "%d", idisplay);
  dpynumlen = strlen(dpynumbuf);
  saddr = NULL;

  /* try hard to get authorization data */
  for (attempt=0; attempt<3; attempt++) {
    if (family == FamilyLocal) {
      if (attempt == 0) {
	saddr = shost->short_name;
	saddrlen = strlen(saddr);
      } else if (attempt == 1) {
	saddr = shost->fq_name;
	saddrlen = strlen(saddr);
      } else {
	break;
      }
    } else if (family == FamilyInternet) {
      if (attempt == 0) {
	saddr = (char *)&shost->addr;
	saddrlen = sizeof(shost->addr);
      } else if (attempt == 1) {
	saddr = shost->fq_name;
	saddrlen = strlen(saddr);
      } else if (attempt == 2) {
	saddr = shost->short_name;
	saddrlen = strlen(saddr);
      } else {
	break;
      }
    } else if (family == FamilyLocalHost) {
      if (attempt == 0) {
	saddr = "localhost";
	saddrlen = strlen(saddr);
      } else {
	break;
      }
    } else {
      break;
    }
#ifdef DEBUG
    if (DEBUGGING(DBG_MISC)) {
      if (saddr == (char *)&shost->addr)
	dprintf("auth", "trying to get authorization with %s", shost->ip_name);
      else
	dprintf("auth", "trying to get authorization with %s", saddr);
    }
#endif
    auth_data = XauGetAuthByAddr(family,
				 (unsigned short) saddrlen, saddr,
				 (unsigned short) dpynumlen, dpynumbuf,
				 (unsigned short) 0, (char *) 0);
    if (auth_data) return(1);
  }
  return(0);
}

/*
 * setup the authorization data so that it can be used by auth_forge()
 */
void auth_setup(void)
{
  if (auth_data == NULL) {
    static Xauth auth_none;
    auth_none.name_length = 0;
    auth_none.name = NULL;
    auth_none.data_length = 0;
    auth_none.data = NULL;
    auth_data = &auth_none;
#ifdef DEBUG
    if (DEBUGGING(DBG_MISC)) dprintf("auth", "no authorization found");
  } else {
    if (DEBUGGING(DBG_MISC)) {
      int i, len = auth_data->data_length;
      char *data = (char *)safe_malloc(2*len+1);
      char *cp = str_ncopy(auth_data->name, auth_data->name_length);
      for (i=0; i<len; i++)
	sprintf(&data[2*i], "%02x", (unsigned char)auth_data->data[i]);
      data[2*len] = '\0';
      dprintf("auth", "found %s authorization of %d bytes: %s", cp, len, data);
      str_free(cp);
      free(data);
    }
#endif
  }
}

/*
 * forge a client connection setup with the authorization found by auth_setup
 * note that buf must contain a valid xConnClientPrefix
 */
MC *auth_forge(char *buf)
{
  MC *mcp;
  U8BITS sex;
  char *cp, *pad = "\0\0\0\0";
  int len, padlen;

  /* Xproto.h/xConnClientPrefix:
     CARD8       byteOrder;
     BYTE        pad;
     CARD16      majorVersion B16, minorVersion B16;
     CARD16      nbytesAuthProto B16;
     CARD16      nbytesAuthString B16;
     CARD16      pad2 B16;
     */
  sex = getCARD8(buf, 0);
  len = 12 + PAD4(auth_data->name_length) + PAD4(auth_data->data_length);
  mcp = mc_new(len);
  /* copy the header and replace the length fields (the rest is intact) */
  memmove(mcp->base, buf, 12);
  putCARD16(sex, &mcp->base[6], auth_data->name_length);
  putCARD16(sex, &mcp->base[8], auth_data->data_length);
  cp = &mcp->base[12];
  /* copy the authorization name */
  len = auth_data->name_length;
  if (len > 0) {
    memmove(cp, auth_data->name, len);
    cp += len;
    padlen = PAD4(len) - len;
    memmove(cp, pad, padlen);
    cp += padlen;
  }
  /* copy the authorization data */
  len = auth_data->data_length;
  if (len > 0) {
    memmove(cp, auth_data->data, len);
    cp += len;
    padlen = PAD4(len) - len;
    memmove(cp, pad, padlen);
    cp += padlen;
  }
  return(mcp);
}
