/*+*****************************************************************************
*                                                                              *
* File: xs.c                                                                   *
*                                                                              *
* Description: X sockets handling                                              *
*                                                                              *
**-****************************************************************************/

#ifndef __lint
char xs_vers[] = "$Id: xs.c,v 1.26 2008/07/11 13:06:57 cons Exp $";
#endif /* __lint */

/******* Include Files ********************************************************/

#include "xs.h"
#include "util.h"
#include "host.h"
#include "auth.h"
#include "err.h"

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <X11/X.h>
#include <X11/Xauth.h>

/******* Constants ************************************************************/

/*
 * from Xproto.h
 */
#define X_TCP_PORT 6000		/* add display number to it */

/*
 * length of the queue for pending connections
 */
#define LISTEN_BACKLOG	10

/*
 * shortcuts because I'm lazy
 */
#define XS_TRY_TCP  (XS_WANT_TCP|XS_NEED_TCP)
#define XS_TRY_UNIX (XS_WANT_UNIX|XS_NEED_UNIX)
#define XS_TRY_LOOP (XS_WANT_LOOP|XS_NEED_LOOP)

/*
 * X server connection type
 */
#define XS_CONNTYPE_UNIX	1	/* UNIX socket */
#define XS_CONNTYPE_LOOP	2	/* TCP socket on loopback interface */
#define XS_CONNTYPE_LOCAL	3	/* TCP socket and X server host is local */
#define XS_CONNTYPE_REMOTE	4	/* TCP socket and X server host is remote */

/******* Macros ***************************************************************/

/******* External Stuff *******************************************************/

int xs_flags = 0;		   /* X socket flags */
int xs_fd[] = { -1, -1, -1 };      /* file descriptors of the incomming sockets */

/******* Local Stuff **********************************************************/

static HOST *xs_host;		/* host of the X server */
static int   xs_dispno;		/* display number of the X server */
static int   xs_conntype;       /* type of connection to the X server */

static char *xs_xsock;		/* path to use for the X sockets */
static char *xs_xdir;		/* directory containing the X sockets */
static char *xs_un_path;	/* path of the UNIX incomming socket */

/*
 * return the string describing the socket type
 */
char *xs_type(int which)
{
  switch (which) {
  case XS_TCP:  return("TCP");  break;
  case XS_UNIX: return("UNIX"); break;
  case XS_LOOP: return("LOOP"); break;
  }
  return("<unknown>");
}

/*
 * close a socket stored in xs_fd[], ignoring any error
 */

static void xs_close_fd(int which)
{
  close(xs_fd[which]);
  xs_fd[which] = -1;
}

/*
 * set the REUSEADDR socket option
 */

static void xs_set_reuseaddr(int fd)
{
#ifdef SO_REUSEADDR
  int res;
  int one = 1;

  res = setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void *)&one, sizeof(one));
  if (res < 0)
    err_report(NULL, "can't setsockopt SO_REUSEADDR: ", ERROR, 0);
#endif /* SO_REUSEADDR */
}

/*
 * set the LINGER socket option
 */

static void xs_set_linger(int fd)
{
#ifdef SO_LINGER
  int res;
  int linger[2] = { 0, 0 };

  res = setsockopt(fd, SOL_SOCKET, SO_LINGER, (void *)linger, sizeof(linger));
  if (res < 0)
    err_report(NULL, "can't setsockopt SO_LINGER: ", ERROR, 0);
#else /* SO_LINGER */
#ifdef SO_DONTLINGER
  int res;

  res = setsockopt(fd, SOL_SOCKET, SO_DONTLINGER, (void *)NULL, 0);
  if (res < 0)
    err_report(NULL, "can't setsockopt SO_DONTLINGER: ", ERROR, 0);
#endif /* SO_DONTLINGER */
#endif /* SO_LINGER */
}

/*
 * make a TCP connection to the X server
 * from Xtranssock.c/TRANS(SocketINETConnect)
 */
static int xs_connect_tcp(HOST *host)
{
  struct sockaddr_in s_in;
  int fd;
  char buf[8];

  memset((char *)&s_in, '\0', sizeof(s_in));
  s_in.sin_family = AF_INET;
  s_in.sin_addr = host->addr;
  s_in.sin_port = htons(X_TCP_PORT + xs_dispno);
  fd = socket(AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    sprintf(buf, "%d", X_TCP_PORT + xs_dispno);
    err_report(NULL, "can't create TCP socket ", host->ip_name,
	       ":", buf, ": ", ERROR, 0);
    return(-1);
  }
#ifdef TCP_NODELAY
  {
    int one = 1;
    setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, &one, sizeof(one));
  }
#endif
  if (connect(fd, (struct sockaddr *)&s_in, sizeof(s_in)) < 0) {
    sprintf(buf, "%d", X_TCP_PORT + xs_dispno);
    err_report(NULL, "can't connect TCP socket ", host->ip_name,
	       ":", buf, ": ", ERROR, 0);
    return(-1);
  }
  return(fd);
}

/*
 * make a UNIX socket connection to the X server
 * from Xtranssock.c/TRANS(SocketUNIXConnect)
 */
static int xs_connect_unix(void)
{
  struct sockaddr_un s_un;
  int fd;

  memset((char *)&s_un, '\0', sizeof(s_un));
  s_un.sun_family = AF_UNIX;
  sprintf(s_un.sun_path, xs_xsock, xs_dispno);
  fd = socket(AF_UNIX, SOCK_STREAM, 0);
  if (fd < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_WARNING))
      dprintf("warn", "can't create UNIX socket %s: %s", s_un.sun_path, ERROR);
#endif
    return(-1);
  }
  if (connect(fd, (struct sockaddr *)&s_un, sizeof(s_un)) < 0) {
#ifdef DEBUG
    if (DEBUGGING(DBG_WARNING))
      dprintf("warn", "can't connect UNIX socket %s: %s", s_un.sun_path, ERROR);
#endif
    return(-1);
  }
  return(fd);
}

/*
 * make a connection to the X server
 */
int xs_connect(void)
{
  int fd = -1;

  switch (xs_conntype) {
  case XS_CONNTYPE_UNIX:
    fd = xs_connect_unix();
    break;
  case XS_CONNTYPE_LOOP:
    fd = xs_connect_tcp(local_host);
    break;
  case XS_CONNTYPE_LOCAL:
    fd = xs_connect_unix();
    if (fd < 0) fd = xs_connect_tcp(xs_host);
    break;
  case XS_CONNTYPE_REMOTE:
    fd = xs_connect_tcp(xs_host);
    break;
  default:
    die("unexpected X server connection type: %d", xs_conntype);
  }
  return(fd);
}

/*
 * accept a TCP connection
 */
static int xs_accept_tcp(int which, HOST **host)
{
  int newfd;
  struct sockaddr_in s_in;
  socklen_t sinlen;
  HOST *newhost;

  /* accept */
  sinlen = sizeof(s_in);
  newfd = accept(xs_fd[which], (void *)&s_in, &sinlen);
  if (newfd < 0) {
    err_report(NULL, xs_type(which), " accept failed: ", ERROR, 0);
    return(-1);
  }

  /* check host */
  newhost = host_create_by_addr(&s_in.sin_addr);
  if (!newhost) {
    err_report(NULL, "accept rejected for ", inet_ntoa(s_in.sin_addr),
	       ": bad or unknown host", 0);
    return(-1);
  }

  /* return */
#ifdef DEBUG
  if (DEBUGGING(DBG_CONN))
    dprintf("conn", "detected %s connection from %s [%s], port %d",
	    xs_type(which), newhost->name, newhost->ip_name, ntohs(s_in.sin_port));
#endif
  *host = newhost;
  return(newfd);
}

/*
 * accept a connection
 */
static int xs_accept_unix(int which)
{
  int newfd;
  struct sockaddr_un s_un;
  socklen_t sunlen;

  /* accept */
  sunlen = sizeof(s_un);
  newfd = accept(xs_fd[which], (void *)&s_un, &sunlen);
  if (newfd < 0) {
    err_report(NULL, xs_type(which), " accept failed: ", ERROR, 0);
    return(-1);
  }

  /* return */
#ifdef DEBUG
  if (DEBUGGING(DBG_CONN))
    dprintf("conn", "detected %s connection", xs_type(which));
#endif
  return(newfd);
}

/*
 * accept a UNIX socket connection
 */
int xs_accept(int which, HOST **host)
{
  int newfd;

  if (which == XS_UNIX) {
    newfd = xs_accept_unix(which);
    *host = this_host;
  } else {
    newfd = xs_accept_tcp(which, host);
  }
  return(newfd);
}

/*
 * create the sockets to act as a fake X server
 * see connection.c and Xtranssock.c
 */
static int xs_listen(int proposed_dispno)
{
  int dispno = proposed_dispno;
  struct sockaddr_in s_in, s_il;
  struct sockaddr_un s_un;
  int res, oldumask;

  oldumask = umask(0);
  goto just_do_it;
try_again_sam:
  if (xs_fd[XS_TCP]  >= 0) xs_close_fd(XS_TCP);
  if (xs_fd[XS_UNIX] >= 0) xs_close_fd(XS_UNIX);
  if (xs_fd[XS_LOOP] >= 0) xs_close_fd(XS_LOOP);
  dispno++;
just_do_it:
  /*
   * try to create the (normal) TCP socket
   */
  if (BITTST(xs_flags, XS_TRY_TCP)) {
    memset((char *)&s_in, '\0', sizeof(s_in));
    s_in.sin_family = AF_INET;
    if (BITTST(xs_flags, XS_ANY))
      s_in.sin_addr.s_addr = htonl(INADDR_ANY);
    else
      memmove(&s_in.sin_addr, &this_host->addr, sizeof(struct in_addr));
    s_in.sin_port = htons(X_TCP_PORT + dispno);
    xs_fd[XS_TCP] = socket(AF_INET, SOCK_STREAM, 0);
    if (xs_fd[XS_TCP] < 0) die("can't create TCP socket: %s", ERROR);
    xs_set_reuseaddr(xs_fd[XS_TCP]);
  }
  /*
   * try to create the UNIX socket
   */
  if (BITTST(xs_flags, XS_TRY_UNIX)) {
    if (mkdir(xs_xdir, 0777) == 0) chmod(xs_xdir, 0777);
    memset((char *)&s_un, '\0', sizeof(s_un));
    s_un.sun_family = AF_UNIX;
    sprintf(s_un.sun_path, xs_xsock, dispno);
    xs_un_path = str_copy(s_un.sun_path);
    xs_fd[XS_UNIX] = socket(AF_UNIX, SOCK_STREAM, 0);
    if (xs_fd[XS_UNIX] < 0) die("can't create UNIX socket: %s", ERROR);
  }
  /*
   * try to create the loopback TCP socket
   */
  if (BITTST(xs_flags, XS_TRY_LOOP)) {
    memset((char *)&s_il, '\0', sizeof(s_il));
    s_il.sin_family = AF_INET;
    memmove(&s_il.sin_addr, &local_host->addr, sizeof(struct in_addr));
    s_il.sin_port = htons(X_TCP_PORT + dispno);
    xs_fd[XS_LOOP] = socket(AF_INET, SOCK_STREAM, 0);
    if (xs_fd[XS_LOOP] < 0) die("can't create LOOP socket: %s", ERROR);
    xs_set_reuseaddr(xs_fd[XS_LOOP]);
  }
  /*
   * try to bind the (normal) TCP socket
   */
  if (BITTST(xs_flags, XS_TRY_TCP)) {
    res = bind(xs_fd[XS_TCP], (void *)&s_in, sizeof(s_in));
    if (res < 0) {
      if ((errno == EADDRINUSE) && BITTST(xs_flags, XS_HUNT))
	goto try_again_sam;
      if (BITTST(xs_flags, XS_NEED_TCP)) {
	die("can't bind TCP socket: %s", ERROR);
      } else {
	err_report(NULL, "can't bind TCP socket: ", ERROR, 0);
	xs_close_fd(XS_TCP);
      }
    }
  }
  /*
   * try to bind the UNIX socket
   */
  if (BITTST(xs_flags, XS_TRY_UNIX)) {
    unlink(xs_un_path);
    res = bind(xs_fd[XS_UNIX], (void *)&s_un, sizeof(s_un));
    if (res < 0) {
      if ((errno == EADDRINUSE) && BITTST(xs_flags, XS_HUNT))
	goto try_again_sam;
      if (BITTST(xs_flags, XS_NEED_UNIX)) {
	die("can't bind UNIX socket: %s", ERROR);
      } else {
	err_report(NULL, "can't bind UNIX socket: ", ERROR, 0);
	xs_close_fd(XS_UNIX);
      }
    }
  }
  /*
   * try to bind the loopback TCP socket
   */
  if (BITTST(xs_flags, XS_TRY_LOOP)) {
    res = bind(xs_fd[XS_LOOP], (void *)&s_il, sizeof(s_il));
    if (res < 0) {
      if ((errno == EADDRINUSE) && BITTST(xs_flags, XS_HUNT))
	goto try_again_sam;
      if (BITTST(xs_flags, XS_NEED_LOOP)) {
	die("can't bind LOOP socket: %s", ERROR);
      } else {
	err_report(NULL, "can't bind LOOP socket: ", ERROR, 0);
	xs_close_fd(XS_LOOP);
      }
    }
  }
  /*
   * do we have what we need now?
   */
  if ((xs_fd[XS_TCP] < 0) && (xs_fd[XS_UNIX] < 0) && (xs_fd[XS_LOOP] < 0))
    die("can't bind any socket!");
  /*
   * try to set additional TCP socket options
   */
  if (xs_fd[XS_TCP] >= 0)  xs_set_linger(xs_fd[XS_TCP]);
  if (xs_fd[XS_LOOP] >= 0) xs_set_linger(xs_fd[XS_LOOP]);
  /*
   * try to listen now
   */
  if (xs_fd[XS_TCP] >= 0) {
    res = listen(xs_fd[XS_TCP], LISTEN_BACKLOG);
    if (res < 0) die("can't listen to TCP socket: %s", ERROR);
  }
  if (xs_fd[XS_UNIX] >= 0) {
    res = listen(xs_fd[XS_UNIX], LISTEN_BACKLOG);
    if (res < 0) die("can't listen to UNIX socket: %s", ERROR);
  }
  if (xs_fd[XS_LOOP] >= 0) {
    res = listen(xs_fd[XS_LOOP], LISTEN_BACKLOG);
    if (res < 0) die("can't listen to LOOP socket: %s", ERROR);
  }
  /*
   * finally return the new display number
   */
  umask(oldumask);
  return(dispno);
}

/*
 * parse the display to setup xs_host, xs_dispno and xs_conntype
 */
static void parse_display(char *dpy)
{
  char *dv, *cp, *dn;

  /* analyze the display, algorithm from XConnDis.c/_XConnectDisplay */
  dv = str_copy(dpy);
  cp = strchr(dv, ':');
  if (cp == NULL) die("bad display (missing colon): %s", dpy);
  if (cp[1] == ':') die("bad display (DECnet is not supported): %s", dpy);
  *cp++ = '\0'; /* end of host part */
  dn = cp;
  cp = strchr(dn, '.');
  if (cp) *cp = '\0'; /* end of number part */
  xs_dispno = atoi(dn);
  if (*dv == '\0' || !strcmp(dv, "unix")) {
    /* local DISPLAY via UNIX socket */
    xs_host = this_host;
    xs_conntype = XS_CONNTYPE_UNIX;
  } else if (!strcmp(dv, "localhost")) {
    /* local DISPLAY via loopback interface */
    xs_host = this_host;
    xs_conntype = XS_CONNTYPE_LOOP;
  } else {
    /* remote DISPLAY */
    xs_host = host_create_by_name(dv);
    if (xs_host == NULL) die("bad display (unknown host %s): %s", dv, dpy);
    if (!strcmp(this_host->ip_name, xs_host->ip_name)) {
      /* same host as this_host */
      host_destroy(xs_host);
      xs_host = this_host;
      xs_conntype = XS_CONNTYPE_LOCAL;
    } else {
      /* real remote DISPLAY */
      xs_conntype = XS_CONNTYPE_REMOTE;
    }
  }
  str_free(dv);
#ifdef DEBUG
  switch (xs_conntype) {
  case XS_CONNTYPE_UNIX:   cp = "unix";     break;
  case XS_CONNTYPE_LOOP:   cp = "loopback"; break;
  case XS_CONNTYPE_LOCAL:  cp = "local";    break;
  case XS_CONNTYPE_REMOTE: cp = "remote";   break;
  default: die("unexpected X server connection type: %d", xs_conntype);
  }
  if (DEBUGGING(DBG_MISC))
    dprintf("misc", "display is %s, using %s X server %d on host %s", dpy,
	    cp, xs_dispno, xs_host->name);
#endif
}

/*
 * create a fake X server and initialise some data in order
 * to be able to connect to the real X server when needed
 */
void xs_setup(char *dpy, int *dispno_ret, char *xsock)
{
  char *cp;

  /*
   * check xs_flags
   */
  if (!BITTST(xs_flags, XS_TRY_TCP|XS_TRY_UNIX|XS_TRY_LOOP))
    die("bad X socket flags: no type defined");

  /*
   * check xsock, see <sys/un.h> for sun_path hard-coded size...
   */
  if (strlen(xsock) > 90) die("bad X socket path (too long): %s", xsock);
  if (!strstr(xsock, "%d")) die("bad X socket path (no %%d): %s", xsock);
  xs_xsock = xsock;
  xs_xdir = str_copy(xsock);
  cp = strrchr(xs_xdir, '/');
  if (cp == NULL) die("bad X socket path (no /): %s", xsock);
  *cp = '\0';
#ifdef DEBUG
  if (DEBUGGING(DBG_MISC))
    dprintf("misc", "X sockets in %s using path %s", xs_xdir, xs_xsock);
#endif

  /*
   * do the remaining setup
   */
  parse_display(dpy);
  switch (xs_conntype) {
  case XS_CONNTYPE_UNIX:
    auth_try(FamilyLocal, xs_host, xs_dispno);
    break;
  case XS_CONNTYPE_LOOP:
    auth_try(FamilyLocal, xs_host, xs_dispno) ||
    auth_try(FamilyLocalHost, xs_host, xs_dispno);
    break;
  case XS_CONNTYPE_LOCAL:
    break;
    auth_try(FamilyLocal, xs_host, xs_dispno) ||
    auth_try(FamilyInternet, xs_host, xs_dispno);
  case XS_CONNTYPE_REMOTE:
    auth_try(FamilyInternet, xs_host, xs_dispno);
    break;
  }
  auth_setup();
  *dispno_ret = xs_listen(*dispno_ret);
#ifdef DEBUG
  if (DEBUGGING(DBG_MISC))
    dprintf("misc", "created sockets for dispno %d: TCP=%d UNIX=%d LOOP=%d",
	    *dispno_ret, xs_fd[XS_TCP], xs_fd[XS_UNIX], xs_fd[XS_LOOP]);
#endif
}

/*
 * close the fake X server
 */
void xs_close(void)
{
  if (xs_fd[XS_TCP]  >= 0) xs_close_fd(XS_TCP);
  if (xs_fd[XS_UNIX] >= 0) xs_close_fd(XS_UNIX);
  if (xs_fd[XS_LOOP] >= 0) xs_close_fd(XS_LOOP);
  unlink(xs_un_path); /* so that it can be reused immediately... */
  str_free(xs_un_path);
}
